//------------------------------------------------------------------------------
// name: The Anti-Compass™
// desc: A tool that takes you in unknowable directions across the wordspace.
// author: Max Jardetzky
//
// based off of poem-randomwalk.ck by Ge Wang (Winter 2023) and kb.ck by Ge Wang
//
//------------------------------------------------------------------------------

0.2 => dac.gain;

/* constants */
10 => int NUM_SEED_WORDS;
15 => int FAST_PRINT_SPEED;
75 => int SLOW_PRINT_SPEED;
10 => int NUM_DIRECTIONAL_WORDS;
1 => int COUNT_TERMINATING_VOWELS;
"\n\n" => string PHRASE_SUFFIX;

/* fun functions to print slowly */
fun void coolPrintln(string s, int wait)
{
    coolPrint(s, wait);
    chout <= IO.newline();
}

fun void coolPrint(string s, int wait)
{
    for (0 => int i; i < s.length(); i++)
    {
        chout <= s.substring(i, 1);
        chout.flush();
        wait::ms => now;
    }
}

ConsoleInput in;
StringTokenizer tok;
string line[0];

/* provides an array of options and returns the index the user chooses */
fun int retrieveInput(string prompt, string options[])
{
    coolPrintln(prompt, FAST_PRINT_SPEED);
    for (0 => int i; i < options.cap(); i++) 
    {
       coolPrintln(i + 1 + ". " + options[i], FAST_PRINT_SPEED);
    }
    coolPrintln("Type 1 through " + options.cap() + " to choose and hit ENTER.", FAST_PRINT_SPEED);

    0 => int chosen;
    while (true)
    {
        in.prompt("") => now;
        
        // read
        while (in.more())
        {
            // get it
            tok.set (in.getLine());
            // clear array
            line.clear();
            // print tokens
            while (tok.more())
            {
                // put into array
                line << tok.next().lower();
            }
            // if non-empty
            if (line.size())
            {
                Std.atoi(line[0]) - 1 => chosen;
                if (chosen >= 0 && chosen < options.cap())
                {
                    return chosen;
                }
                coolPrintln("Invalid option, try again!", FAST_PRINT_SPEED);
            }
        }
    }
}

// instantiate
Word2Vec model;

// pre-trained model to load
"glove-wiki-gigaword-50-tsne-2-filtered.txt" => string filepath;

<<< "Loading model (this could take a few seconds)...", "" >>>;

// load a pre-trained model (see URLs above for download)
if (!model.load(me.dir() + filepath))
{
    <<< "Cannot load model:", filepath >>>;
    me.exit();
}

// ranges for each dimension (for sound mapping)
float mins[0], maxs[0];

// get bounds for each dimension
model.minMax(mins, maxs);

// print info
// print(model);

// assemble N random words from randomly generated coordinates
string seed_words[NUM_SEED_WORDS];
for (0 => int i; i < NUM_SEED_WORDS; i++)
{
    // set up float vector for word
    float vec[model.dim()];
    
    // generate random coordinate along each dimension
    for (0 => int j; j < model.dim(); j++)
    {
        Math.random2f(mins[j], maxs[j]) => vec[j];
    }

    // add the nearest word to the seed words array
    string result[1];
    model.getSimilar(vec, 1, result);
    result[0] => seed_words[i];
}

// intro segment
coolPrint("Thank you for choosing the ", FAST_PRINT_SPEED);

Event ev1;
Event ev2;

spork ~ sonify(ev1, "anti");
spork ~ coolPrint("Anti-", SLOW_PRINT_SPEED);
ev1 => now;
spork ~ sonify(ev2, "compass");
spork ~ coolPrint("Compass™", SLOW_PRINT_SPEED);
ev2 => now;

retrieveInput(". The Anti-Compass™ has provided you with the following words to begin your journey. From then on, you will either be able to choose the direction or magnitude by which to traverse wordspace, and the Anti-Compass™ will assist you.", seed_words) => int desired;

/* function to count the number of vowels in a string */
fun int numVowels(string word)
{
    word.lower() => word;
    0 => int vowelCount;
    for (0 => int i; i < word.length(); i++)
    {
        word.substring(i, 1) => string letter;
        if (letter == "a" || letter == "e" || letter == "i" || letter == "o" || letter == "u" || letter == "y")
        {
            // only count it if the previous letter wasn't a vowel or it's not a terminating vowel
            if (COUNT_TERMINATING_VOWELS || i < word.length() - 1)
            {
                if (i > 0)
                {
                    word.substring(i-1, 1) => string priorLetter;
                    if (priorLetter != "a" && priorLetter != "e" && priorLetter != "i" && priorLetter != "o" && priorLetter != "u" && priorLetter != "y")
                    {
                        vowelCount++;
                    }
                }
                else
                {
                    vowelCount++;
                }
            }
        }
    }
    return vowelCount;
}

/* function to weight the average position of letters in the alphabet from 0 to 1 */
fun float avgPosition(string word)
{
    word.lower() => word;
    0 => float alphabetSum;
    for (0 => int i; i < word.length(); i++)
    {
        word.charAt(i) - 96 => int charNum;
        if (charNum >= 1 && charNum <= 26) charNum +=> alphabetSum;  // only count if letter to prevent glitches
    }
    return alphabetSum / (26 * word.length());
}

/* function to sonify a word vector, outputting to dac */
fun void sonify(Event ev, string word)
{
    if (word == PHRASE_SUFFIX)
    {
        1::second => now;
        ev.signal();
        return;
    }
    float vec[model.dim()];
    model.getVector(word, vec);

    BeeThree one => Echo e => NRev r => dac;
    Moog two => e;
    0.75 => one.gain;
    // delay and reverb constants (high delay/no reverb at min y and high reverb/no delay at max y)
    10::ms => e.delay;
    // if y is negative, starting pitch is mapped from C4 to C5 and goes downward by number of consonants
    // if y is positive, starting pitch is mapped from C2 to C3 and goes upward by number of consonants
    (36 + (vec[0] - mins[0]) / (maxs[0] - mins[0]) * 12) $ int => int midiPitch;
    word.length() - numVowels(word) => int midiDelta;
    if (vec[1] < 0)
    {
        108 - midiPitch => midiPitch;
        -1 *=> midiDelta;
    }
    // length of strikes and distance between strikes is shorter as average letter weight is closer to a
    avgPosition(word) => float avgLetterPos;
    // the number of vowels in the word (unless one vowel) is the number of hits
    numVowels(word) > 1 ? numVowels(word) : 1 => int numHits;
    for (0 => int i; i < numHits; i++)
    {
        // configure delay and reverb
        Math.fabs((vec[1] - mins[1]) / (maxs[1] - mins[1])) * 0.25 => e.mix;
        0.25 - e.mix() => r.mix;

        // set frequencies (first cap midiPitch between C1 and C7)
        if (midiPitch < 24)
        {
            24 => midiPitch;
        }
        else if (midiPitch > 96)
        {
            96 => midiPitch;
        }
        midiPitch - 12 => Std.mtof => one.freq;
        midiPitch => Std.mtof => two.freq;

        // trigger notes
        one.noteOn(0.8);
        two.noteOn(0.8);

        // strike duration
        (10 + 40 * avgLetterPos)::ms => now;

        // end notes
        one.noteOff(0.8);
        two.noteOff(0.8);

        // change pitch
        midiDelta +=> midiPitch;

        // wait for next strike
        (75 + 175 * avgLetterPos)::ms => now;
    }
    
    // allow reverb tails (halve reverb and delay many times over time to fade out)
    100::ms => now;
    for (0 => int i; i < 500; i++)
    {
        e.mix() / 1.01 => e.mix;
        r.mix() / 1.01 => r.mix;
        5::ms => now;

        if (i == 80)
        {
            // signal after 0.5 second
            ev.signal();
        }
    }
}

// insurance policy in case sonify takes 0.5 seconds to signal but we don't finish printing until 1 second (lost signal)
fun void insurance(Event ev)
{
    2::second => now;
    ev.signal();
}

/* working variables for game */
["pick cardinal direction (forgo magnitude)", "pick magnitude (forgo cardinal direction)", "new phrase", "replay", "conclude"] @=> string initOptions[];
["north", "northeast", "east", "southeast", "south", "southwest", "west", "northwest"] @=> string directions[];
seed_words[desired] => string word;
word => string lastWord;
string wordBuf[1024];
0 => int curWord;
int initOption;
int direction;
int chosenOption;
"You have chosen the word " => string wordChoicePrefix;
0 => int replayed;

// for later use
fun void replay()
{
    // repeat and replay the stored words
    for (0 => int i; i < curWord; i++)
    {
        Event ev;
        spork ~ sonify(ev, wordBuf[i]);
        coolPrint(wordBuf[i], SLOW_PRINT_SPEED);
        if (wordBuf[i] != PHRASE_SUFFIX)
        {
            coolPrint(" ", SLOW_PRINT_SPEED);
        }
        spork ~ insurance(ev);
        ev => now;
    }
    coolPrintln("", FAST_PRINT_SPEED);
}

// play the game and fill out the word buffer
while (true)
{
    // add word to textual buffer, only if we didn't replay
    if (replayed == 0) word => wordBuf[curWord++];

    // print and sonify the word
    if (word == PHRASE_SUFFIX)
    {
        coolPrintln("You have chosen to begin a new phrase.", FAST_PRINT_SPEED);
        lastWord => word;
    }
    else
    {
        Event ev;
        coolPrint(wordChoicePrefix, FAST_PRINT_SPEED);
        spork ~ sonify(ev, word);
        coolPrint(word + ".", SLOW_PRINT_SPEED);
        spork ~ insurance(ev);
        ev => now;
        coolPrintln("", FAST_PRINT_SPEED);
    }

    // choose between direction or magnitude.
    retrieveInput("Would you like to pick the cardinal direction, choose a magnitude, begin a new phrase, replay the current state, or conclude?", initOptions) => initOption;
    if (initOption == 0)  // pick direction, forgo magnitude
    {
        Math.random2(0, NUM_DIRECTIONAL_WORDS - 1) => chosenOption;
    }
    if (initOption == 1)  // pick magnitude, forgo direction
    {
        Math.random2(0, directions.cap() - 1) => direction;
    }
    else if (initOption == 2)  // new phrase
    {
        word => lastWord;
        PHRASE_SUFFIX => word;
        0 => replayed;
        continue;
    }
    else if (initOption == 3) // replay
    {
        replay();
        1 => replayed;
        continue;
    }
    else if (initOption == 4)  // conclude
    {
        break;
    }

    // choose direction for next word
    if (initOption == 0)
    {
        retrieveInput("Now, choose the cardinal direction in the wordspace in which to explore for the next word.", directions) => direction;
    }

    // handle directional logic (extend halfway to the end of the vector space in the direction => average)
    float startVec[model.dim()];
    float endVec[model.dim()];
    for (0 => int i; i < model.dim(); i++)
    {
        startVec[i] => endVec[i];  // deep copy
    }
    string nextOptions[NUM_DIRECTIONAL_WORDS];

    model.getVector(word, startVec);
    if (direction == 0)  // north (zero x, positive y)
    {
        (endVec[1] + maxs[1]) / 2 => endVec[1];
    }
    else if (direction == 1)  // northeast (positive x, positive y)
    {
        (endVec[0] + maxs[0]) / 2 => endVec[0];
        (endVec[1] + maxs[1]) / 2 => endVec[1];
    }
    else if (direction == 2)  // east (positive x, zero y)
    {
        (endVec[0] + maxs[0]) / 2 => endVec[0];
    }
    else if (direction == 3)  // southeast (positive x, negative y)
    {
        (endVec[0] + maxs[0]) / 2 => endVec[0];
        (endVec[1] + mins[1]) / 2 => endVec[1];
    }
    else if (direction == 4)  // south (zero x, negative y)
    {
        (endVec[1] + mins[1]) / 2 => endVec[1];
    }
    else if (direction == 5)  // southwest (negative x, negative y)
    {
        (endVec[0] + mins[0]) / 2 => endVec[0];
        (endVec[1] + mins[1]) / 2 => endVec[1];
    }
    else if (direction == 6)  // west (negative x, zero y)
    {
        (endVec[0] + mins[0]) / 2 => endVec[0];
    }
    else if (direction == 7)  // northwest (negative x, positive y)
    {
        (endVec[0] + mins[0]) / 2 => endVec[0];
        (endVec[1] + maxs[1]) / 2 => endVec[1];
    }

    initOption == 1 ? "The Anti-Compass™ has taken you " : "You have gone " => string directionalPrefix;
    coolPrintln(directionalPrefix + directions[direction] + ".", FAST_PRINT_SPEED);

    // divide space into chunks between startVec and endVec and present the nearest words for the coordinates
    string directionOptions[NUM_DIRECTIONAL_WORDS];
    (endVec[0] - startVec[0]) / NUM_DIRECTIONAL_WORDS => float dx;
    (endVec[1] - startVec[1]) / NUM_DIRECTIONAL_WORDS => float dy;
    for (1 => int i; i <= NUM_DIRECTIONAL_WORDS; i++)
    {
        // deep copy base vector in case of higher dimensions
        float tempVec[model.dim()];
        for (0 => int j; j < model.dim(); j++)
        {
            startVec[j] => tempVec[j];
        }

        // compute and add delta in first two dimensions
        startVec[0] + i * dx => tempVec[0];
        startVec[1] + i * dy => tempVec[1];

        // get most similar word and store in the return array
        string result[1];
        model.getSimilar(tempVec, 1, result);
        result[0] => directionOptions[i - 1];
    }

    initOption == 1 ? "You have chosen the word " : "The Anti-Compass™ has chosen the word " => wordChoicePrefix;
    if (initOption == 1)
    {
        retrieveInput("Choose from the following words, ordered in terms of linearly decreasing similarity in the wordspace along your chosen direction.", directionOptions) => chosenOption;
    }
    directionOptions[chosenOption] => word;
    0 => replayed;
}

replay();
1::second => now;

// print info (for fun and debugging)
fun void print( Word2Vec @ m )
{
    <<< "dictionary size:", m.size() >>>;
    <<< "embedding dimensions:", m.dim() >>>;
    <<< "using KDTree for search:", m.useKDTree() ? "YES" : "NO" >>>;
    // ranges for each dimension (for sound mapping)
    float mins[0], maxs[0];
    // get bounds for each dimension
    model.minMax( mins, maxs );
    // print the min/max for each dimension
    for( int i; i < m.dim(); i++ )
    { cherr <= "dimension " <= i <= ": [" <= mins[i]
    <= " <-range-> " <= maxs[i] <= "]" <= IO.newline(); }
}
